/*
 * Copyright (C) 1995 M. Hauber, Ch. Schneider, G. Caronni
 * See COPYING for more details
 */

/*
 * This code implements the MD5 message-digest algorithm.
 * The algorithm is due to Ron Rivest.  This code was
 * written by Colin Plumb in 1993, no copyright is claimed.
 * This code is in the public domain; do with it what you wish.
 *
 * Equivalent code is available from RSA Data Security, Inc.
 * This code has been tested against that, and is equivalent,
 * except that you don't need to include two pages of legalese
 * with every copy.
 *
 * To compute the message digest of a chunk of bytes, declare an
 * MD5_Context structure, pass it to MD5Init, call MD5Update as
 * needed on buffers full of bytes, and then call MD5Final, which
 * will fill a supplied 16-byte array with the digest.
 *
 * NOTE: Modified for SKIP to be close to RSA implementation (cschneid)
 *       - Type MD5_CTX instead of struct MD5_Context
 *       - MD5_CTX.digest is filled on MD5Final, user needs to copy manually
 */

#include "config.h"
#include "memblk.h"
#include "sign.h"
#include "signmd5.h"

#ifdef __GNUC__
#ident "$Id: signmd5.c,v 1.7 1996/05/23 13:21:17 cschneid Exp $"
#else
static char rcsid[] = "$Id: signmd5.c,v 1.7 1996/05/23 13:21:17 cschneid Exp $";
#endif

#ifndef HIGHFIRST
#define byteReverse(buf, len)	/* Nothing */
#else
static void byteReverse(unsigned char *buf, unsigned longs);

#ifndef ASM_MD5
/*
 * Note: this code is harmless on little-endian machines.
 */
static void byteReverse(unsigned char *buf, unsigned longs)
{
    u_int32 t;
    do {
	t = (u_int32) ((unsigned) buf[3] << 8 | buf[2]) << 16 |
	    ((unsigned) buf[1] << 8 | buf[0]);
	*(u_int32 *) buf = t;
	buf += 4;
    } while (--longs);
}
#endif
#endif

static void MD5Transform(u_int32 buf[4], u_int32 const in[16]);

/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
void MD5Init(struct MD5_CTX *ctx)
{
    ctx->buf[0] = 0x67452301;
    ctx->buf[1] = 0xefcdab89;
    ctx->buf[2] = 0x98badcfe;
    ctx->buf[3] = 0x10325476;

    ctx->bits[0] = 0;
    ctx->bits[1] = 0;
}

/*
 * Update context to reflect the concatenation of another buffer full
 * of bytes.
 */
void MD5Update(struct MD5_CTX *ctx, unsigned char const *buf, 
								unsigned len)
{
    u_int32 t;

    /* Update bitcount */

    t = ctx->bits[0];
    if ((ctx->bits[0] = t + ((u_int32) len << 3)) < t)
	ctx->bits[1]++;		/* Carry from low to high */
    ctx->bits[1] += len >> 29;

    t = (t >> 3) & 0x3f;	/* Bytes already in shsInfo->data */

    /* Handle any leading odd-sized chunks */

    if (t) {
	unsigned char *p = (unsigned char *) ctx->in + t;

	t = 64 - t;
	if (len < t) {
	    MEMCPY(p, buf, len);
	    return;
	}
	MEMCPY(p, buf, t);
	byteReverse(ctx->in, 16);
	MD5Transform(ctx->buf, (u_int32 *) ctx->in);
	buf += t;
	len -= t;
    }
    /* Process data in 64-byte chunks */

    while (len >= 64) {
	MEMCPY(ctx->in, buf, 64);
	byteReverse(ctx->in, 16);
	MD5Transform(ctx->buf, (u_int32 *) ctx->in);
	buf += 64;
	len -= 64;
    }

    /* Handle any remaining bytes of data. */

    MEMCPY(ctx->in, buf, len);
}

/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, MSB-first)
 */
void MD5Final(struct MD5_CTX *ctx)
{
    unsigned count;
    unsigned char *p;

    /* Compute number of bytes mod 64 */
    count = (ctx->bits[0] >> 3) & 0x3F;

    /* Set the first char of padding to 0x80.  This is safe since there is
       always at least one byte free */
    p = ctx->in + count;
    *p++ = 0x80;

    /* Bytes of padding needed to make 64 bytes */
    count = 64 - 1 - count;

    /* Pad out to 56 mod 64 */
    if (count < 8) {
	/* Two lots of padding:  Pad the first block to 64 bytes */
	MEMZERO(p, count);
	byteReverse(ctx->in, 16);
	MD5Transform(ctx->buf, (u_int32 *) ctx->in);

	/* Now fill the next block with 56 bytes */
	MEMZERO(ctx->in, 56);
    } else {
	/* Pad block to 56 bytes */
	MEMZERO(p, count - 8);
    }
    byteReverse(ctx->in, 14);

    /* Append length in bits and transform */
    ((u_int32 *) ctx->in)[14] = ctx->bits[0];
    ((u_int32 *) ctx->in)[15] = ctx->bits[1];

    MD5Transform(ctx->buf, (u_int32 *) ctx->in);
    byteReverse((unsigned char *) ctx->buf, 4);
    MEMCPY(ctx->digest, ctx->buf, sizeof(ctx->digest));
    /* Wipe out (possibly) sensitive data */
    MEMZERO(ctx->buf, sizeof(ctx->buf));
    MEMZERO(ctx->bits, sizeof(ctx->bits));
    MEMZERO(ctx->in, sizeof(ctx->in));
}

#ifndef ASM_MD5

/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f, w, x, y, z, data, s) \
	( w += f(x, y, z) + data,  w = w<<s | w>>(32-s),  w += x )

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
static void MD5Transform(u_int32 buf[4], u_int32 const in[16])
{
    register u_int32 a, b, c, d;

    a = buf[0];
    b = buf[1];
    c = buf[2];
    d = buf[3];

    MD5STEP(F1, a, b, c, d, in[0] + 0xd76aa478, 7);
    MD5STEP(F1, d, a, b, c, in[1] + 0xe8c7b756, 12);
    MD5STEP(F1, c, d, a, b, in[2] + 0x242070db, 17);
    MD5STEP(F1, b, c, d, a, in[3] + 0xc1bdceee, 22);
    MD5STEP(F1, a, b, c, d, in[4] + 0xf57c0faf, 7);
    MD5STEP(F1, d, a, b, c, in[5] + 0x4787c62a, 12);
    MD5STEP(F1, c, d, a, b, in[6] + 0xa8304613, 17);
    MD5STEP(F1, b, c, d, a, in[7] + 0xfd469501, 22);
    MD5STEP(F1, a, b, c, d, in[8] + 0x698098d8, 7);
    MD5STEP(F1, d, a, b, c, in[9] + 0x8b44f7af, 12);
    MD5STEP(F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
    MD5STEP(F1, b, c, d, a, in[11] + 0x895cd7be, 22);
    MD5STEP(F1, a, b, c, d, in[12] + 0x6b901122, 7);
    MD5STEP(F1, d, a, b, c, in[13] + 0xfd987193, 12);
    MD5STEP(F1, c, d, a, b, in[14] + 0xa679438e, 17);
    MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

    MD5STEP(F2, a, b, c, d, in[1] + 0xf61e2562, 5);
    MD5STEP(F2, d, a, b, c, in[6] + 0xc040b340, 9);
    MD5STEP(F2, c, d, a, b, in[11] + 0x265e5a51, 14);
    MD5STEP(F2, b, c, d, a, in[0] + 0xe9b6c7aa, 20);
    MD5STEP(F2, a, b, c, d, in[5] + 0xd62f105d, 5);
    MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
    MD5STEP(F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
    MD5STEP(F2, b, c, d, a, in[4] + 0xe7d3fbc8, 20);
    MD5STEP(F2, a, b, c, d, in[9] + 0x21e1cde6, 5);
    MD5STEP(F2, d, a, b, c, in[14] + 0xc33707d6, 9);
    MD5STEP(F2, c, d, a, b, in[3] + 0xf4d50d87, 14);
    MD5STEP(F2, b, c, d, a, in[8] + 0x455a14ed, 20);
    MD5STEP(F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
    MD5STEP(F2, d, a, b, c, in[2] + 0xfcefa3f8, 9);
    MD5STEP(F2, c, d, a, b, in[7] + 0x676f02d9, 14);
    MD5STEP(F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

    MD5STEP(F3, a, b, c, d, in[5] + 0xfffa3942, 4);
    MD5STEP(F3, d, a, b, c, in[8] + 0x8771f681, 11);
    MD5STEP(F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
    MD5STEP(F3, b, c, d, a, in[14] + 0xfde5380c, 23);
    MD5STEP(F3, a, b, c, d, in[1] + 0xa4beea44, 4);
    MD5STEP(F3, d, a, b, c, in[4] + 0x4bdecfa9, 11);
    MD5STEP(F3, c, d, a, b, in[7] + 0xf6bb4b60, 16);
    MD5STEP(F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
    MD5STEP(F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
    MD5STEP(F3, d, a, b, c, in[0] + 0xeaa127fa, 11);
    MD5STEP(F3, c, d, a, b, in[3] + 0xd4ef3085, 16);
    MD5STEP(F3, b, c, d, a, in[6] + 0x04881d05, 23);
    MD5STEP(F3, a, b, c, d, in[9] + 0xd9d4d039, 4);
    MD5STEP(F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
    MD5STEP(F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
    MD5STEP(F3, b, c, d, a, in[2] + 0xc4ac5665, 23);

    MD5STEP(F4, a, b, c, d, in[0] + 0xf4292244, 6);
    MD5STEP(F4, d, a, b, c, in[7] + 0x432aff97, 10);
    MD5STEP(F4, c, d, a, b, in[14] + 0xab9423a7, 15);
    MD5STEP(F4, b, c, d, a, in[5] + 0xfc93a039, 21);
    MD5STEP(F4, a, b, c, d, in[12] + 0x655b59c3, 6);
    MD5STEP(F4, d, a, b, c, in[3] + 0x8f0ccc92, 10);
    MD5STEP(F4, c, d, a, b, in[10] + 0xffeff47d, 15);
    MD5STEP(F4, b, c, d, a, in[1] + 0x85845dd1, 21);
    MD5STEP(F4, a, b, c, d, in[8] + 0x6fa87e4f, 6);
    MD5STEP(F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
    MD5STEP(F4, c, d, a, b, in[6] + 0xa3014314, 15);
    MD5STEP(F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
    MD5STEP(F4, a, b, c, d, in[4] + 0xf7537e82, 6);
    MD5STEP(F4, d, a, b, c, in[11] + 0xbd3af235, 10);
    MD5STEP(F4, c, d, a, b, in[2] + 0x2ad7d2bb, 15);
    MD5STEP(F4, b, c, d, a, in[9] + 0xeb86d391, 21);

    buf[0] += a;
    buf[1] += b;
    buf[2] += c;
    buf[3] += d;
}

#endif

/*
 * skip sign algorithm interface to md5
 */
static int signmd5_init(void)
{
  return 0;
}

static int signmd5_exit(void)
{
  return 0;
}

static int signmd5_icvlen(void)
{
  MD5_CTX x;

  return sizeof(x.digest);
}


static int signmd5_sign(u_char *Kp, int Kplen, struct memblk *m, u_char *icv)
{
  int len;
  MD5_CTX ctx;

  MD5Init(&ctx);
  /* 
   * Keyed-MD5 conforming to Internet Draft draft-krawczyk-keyed-md5-00.txt
   * There are faster ways to do this... (See draft)
   */
  MD5Update(&ctx, Kp, Kplen);  /* Prepend padded Kp to message */
  if (Kplen < 64)
  {
    static u_char Kppadding[64];
    static unsigned long zilch;
    static unsigned long lolen;
    static u_char fck;
    Kppadding[0] = 0x80;

#if 0
    MD5Update(&ctx, Kppadding, 64 - Kplen);
XXX this is the wrong way to do padding -- you have to add a length field
in the last 8 bytes. if it does not fit in the last 8 bytes, you need to fill
another pad block, and put it there.
the length field looks like
    | lsbyte |  ...  | msbyte |

all in one would be more efficient

#else

if (64-Kplen<8) printf("(64-Kplen<8) you should panic now...\n");

    MD5Update(&ctx, Kppadding, 64 - Kplen - 8);
    lolen=Kplen*8;
    fck=lolen&0xff; lolen = lolen >> 8; MD5Update(&ctx, &fck, 1);
    fck=lolen&0xff; lolen = lolen >> 8; MD5Update(&ctx, &fck, 1);
    fck=lolen&0xff; lolen = lolen >> 8; MD5Update(&ctx, &fck, 1);
    fck=lolen&0xff;                     MD5Update(&ctx, &fck, 1);
    MD5Update(&ctx, &zilch, 4);
#endif

  }

  while (m->len > 0)
  {
    len = m->ms->len - m->offset;

    if (len > m->len)
      len = m->len;

    MD5Update(&ctx, BLKSTART(m), len);
    if (((m->len -= len) > 0) || (len == (m->ms->len - m->offset)))
    {
      /* Not last segment OR last segment fully used */
      m->ms++;
      m->offset = 0;
    }
    else
    {
      /* Last segment only partially used */
      m->offset += len;
    }
  }

  MD5Update(&ctx, Kp, Kplen);  /* Append Kp to message */
  MD5Final(&ctx);
  MEMCPY(icv, ctx.digest, sizeof(ctx.digest));

  return 0;
}

/* Information structure containing pointers to algorithm functions */
struct sign_algorithm signmd5 =
{
  signmd5_init, signmd5_exit,
  signmd5_sign,
  signmd5_icvlen
};
